"""Provides tools to evaluate Spark models in mlpype."""
from typing import Dict, List, Optional, Union

from pyspark.ml.evaluation import JavaEvaluator
from pyspark.sql import DataFrame as SparkDataFrame

from mlpype.base.data.dataset import DataSet
from mlpype.base.evaluate.base_evaluator import BaseEvaluator
from mlpype.base.pipeline import Pipeline
from mlpype.spark.model import SparkModel


class SparkEvaluator(BaseEvaluator):
    """Used to evaluate Spark models in mlpype."""

    def __init__(
        self,
        evaluator: JavaEvaluator,
        metrics: List[str],
        set_cache: bool = False,
    ) -> None:
        """Used to evaluate Spark models in mlpype.

        Args:
            evaluator (JavaEvaluator): One of the Evaluators from pyspark, e.g.
                psypark.ml.evaluation.RegressionEvaluator
            metrics (List[str]): The list of metrics to retrieve using the Evaluator.
                See the documentation of the one you use to know which are available.
            set_cache (bool, optional): If True, the predictions will be cached just before
                evaluation. This can help speed up the evaluation. Defaults to False.
        """
        super().__init__()
        self.evaluator = evaluator
        self.metrics = metrics
        self.set_cache = set_cache

    def evaluate(
        self, model: SparkModel, data: DataSet, pipeline: Optional[Pipeline] = None
    ) -> Dict[str, Union[str, float, int, str, bool]]:
        """Evaluate the given model on the given dataset.

        We assume the model does not need to be transformed anymore if pipeline is None.

        Args:
            model (SparkModel): The SparkModel to evaluate.
            data (DataSet): The Dataset to use to evaluate the model.
            pipeline (Optional[Pipeline]): If not None, this will be used to transform `data` first.

        Returns:
            Dict[str, Union[str, float, int, str, bool]]: A dictionary of metric_name-value pairs. The result
                of the evaluation.
        """
        if pipeline is not None:
            data = pipeline.transform(data)

        predictions = model.transform(data).get_all(model.outputs)

        assert len(predictions) == 1, "Expect exactly 1 output from a SparkModel."
        preds: SparkDataFrame = predictions[0]
        if self.set_cache:
            preds = preds.cache()

        result = {}
        for metric in self.metrics:
            value = self.evaluator.setMetricName(metric).evaluate(preds)  # type: ignore
            result[metric] = value
        return result

    def __str__(self, indents: int = 0) -> str:
        """Generates a string representation of this SparkEvaluator.

        Args:
            indents (int, optional): The number of tabs to preceed this string
                representation. Defaults to 0.

        Returns:
            str: A string representation of this SparkEvaluator.
        """
        tabs = "\t" * indents
        metrics = ", ".join(self.metrics)
        return f"{tabs}{type(self.evaluator).__name__}: {metrics}"



"""Provides tools for using Sklearn with MLpype.

This provides a few tools:

1. An SklearnModel: An MLpype model that uses Sklearn. This is easily extended to most models in Sklearn.
This can be found in `mlpype.sklearn.model`. This module also contains pre-loaded models from Sklearn.
2. Type checkers for Numpy and Pandas data.

MLpype base pipelines/pipes should already be compatible with Sklearn transformers, so you can use them directly.
StandardScaler for instance will work with MLpype.
"""
from . import data, model, pipeline



"""Provides an integration layer between MLpype-trained models and MLflow models."""
from pathlib import Path

from mlflow.pyfunc import PythonModel, PythonModelContext

from mlpype.base.data.dataset import DataSet
from mlpype.base.deploy.inference import Inferencer


class PypeMLFlowModel(PythonModel):
    """An integration layer between MLpype-trained models and MLflow models."""

    def load_context(self, context: PythonModelContext) -> None:
        """
        Loads the context for the Python model.

        This load an Inferencer from the logged folder.

        Parameters:
            context (PythonModelContext): The default context provided by MLpype + MLflow.
                This included 1 variable, the path to the logged folder.
                This is the lower level artifact folder: MLPype stores
                the results higher up in the folder structure.
        """
        assert "folder" in context.artifacts
        root = Path(context.artifacts["folder"])
        print(f"Starting in {root}")
        for _ in range(4):
            if (root / "extra_files.json").exists():
                break
            root = root.parent
            print(f"Trying {root}")
        else:
            raise ValueError("Could not locate proper root folder to import model.")
        self.inferencer = Inferencer.from_folder(root)

    def predict(self, context: PythonModelContext, model_input: DataSet) -> DataSet:
        """
        Makes a predition using the loaded MLpype Inferencer.

        Args:
            context (PythonModelContext): The context for the Python model. Not used,
                but required as a default argument.
            model_input (DataSet): The input data for the model prediction.

        Returns:
            DataSet: The predicted output dataset.
        """
        return self.inferencer.predict(model_input)



"""A mlpype implementation of lightgbm using the Sklearn interface."""

from lightgbm import LGBMModel

from mlpype.sklearn.model.sklearn_model import SklearnModel


class LightGBMModel(SklearnModel[LGBMModel]):
    """A mlpype implementation of lightgbm using the Sklearn interface.

    For Mac users: please keep in mind that OpenMP may not be installed
    correctly for lightgbm to use it. If your training hangs, set
    the number of threads to 1.
    """



"""Provides tools for using Spark DataFrames as an input data source."""
from .spark_data_frame_source import SparkDataFrameSource
from .spark_read import SparkReadSource
from .spark_sql_source import SparkSqlSource



"""Provides a generic class for sklearn-like Models."""
from typing import Any, Iterable, Protocol, Union

from mlpype.sklearn.data.sklearn_data import SklearnData


class SklearnModelBaseType(Protocol):
    """Base class for sklearn-like models."""

    def fit(self, *x: SklearnData, **kwargs: Any) -> Any:
        """Fit a model to the given data. Kwargs are ignored."""

    def predict(self, *x: SklearnData, **kwargs: Any) -> Union[Iterable[SklearnData], SklearnData]:
        """Predict for given data using a trained model. Kwargs are ignored."""



"""Provides wrapper classes for serialising and deserialising Spark components."""
from pyspark.ml import Model as BaseSparkModel
from pyspark.ml import Predictor, Transformer
from pyspark.ml.util import MLReader, MLWritable


class SerialisablePredictor(Predictor, MLWritable, MLReader):
    """A Spark Predictor that can be serialised and deserialised."""


class SerialisableTransformer(Transformer, MLWritable, MLReader):
    """A Spark Transformer that can be serialised and deserialised."""


class SerialisableSparkModel(BaseSparkModel, MLWritable, MLReader):
    """A Spark Model that can be serialised and deserialised."""



"""Provides a generic class for sklearn-like Models."""
import types
import typing
from argparse import ArgumentParser
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, Generic, Iterable, List, Optional, Type, TypeVar, Union

import numpy as np

from mlpype.base.experiment.argument_parsing import add_args_to_parser_for_class
from mlpype.base.model import Model
from mlpype.base.serialiser.joblib_serialiser import JoblibSerialiser
from mlpype.sklearn.data.sklearn_data import SklearnData
from mlpype.sklearn.model.sklearn_base_type import SklearnModelBaseType

T = TypeVar("T", bound=SklearnModelBaseType)
logger = getLogger(__name__)


class SklearnModel(Model[SklearnData], Generic[T]):
    """A generic class for sklearn-like Models.

    You should set a sklearn model as a type hint to this class when defining a new model.
    This allows us to get the parameters from the documentation of that sklearn model.
    For an example, see the implementation of LinearModel, especially the `SklearnModel[LinearRegression]` part.

    Below are some examples for how to do this yourself.

    ```python
    # Works
    class LinearRegressionModel(SklearnModel[LinearRegression]):
        pass

    # An alternative to dynamically generate the model, which is easier to export/import
    model_class = SklearnModel.class_from_sklearn_model_class(LinearRegression)

    # Unfortunately, using something like the following will not work due to how Generic types are handled.
    LinearRegressionModel = SklearnModel[LinearRegression]
    ```
    """

    SKLEARN_MODEL_FILE = "model.pkl"

    def __init__(
        self,
        inputs: List[str],
        outputs: List[str],
        model: Optional[T] = None,
        seed: int = 1,
        **model_args: Any,
    ) -> None:
        """A generic class for sklearn-like Models.

        Args:
            inputs (List[str]): A list of names of input Data. This determines which Data is
                used to fit the model.
            outputs (List[str]): A list of names of output Data. This determines the names of
                output variables.
            model (Optional[T]): An object that has fit() and predict() methods. If none,
                we will use the model_args to instantiate a new model. Should be of type SklearnModelBaseType
            seed (int, optional): The RNG seed to ensure reproducability.. Defaults to 1.
            **model_args (Any): Optional keyword arguments passed to the model class to instantiate a new
                model if `model` is None. This is where the arguments to the sklearn model go.
        """
        super().__init__(inputs, outputs, seed)
        if model is None:
            model = self._init_model(model_args)
        self.model = model

    def _init_model(self, args: Dict[str, Any]) -> T:
        return self._get_annotated_class()(**args)

    @classmethod
    def _get_annotated_class(cls) -> Type[T]:
        return typing.get_args(cls.__orig_bases__[0])[0]

    def set_seed(self) -> None:
        """Sets the RNG seed."""
        np.random.seed(self.seed)

    def _fit(self, *data: SklearnData) -> None:
        self.model.fit(*data)

    def _transform(self, *data: SklearnData) -> Union[Iterable[SklearnData], SklearnData]:
        return self.model.predict(*data)

    def _save(self, folder: Path) -> None:
        serialiser = JoblibSerialiser()
        serialiser.serialise(self.model, folder / self.SKLEARN_MODEL_FILE)

    @classmethod
    def _load(cls: Type["SklearnModel"], folder: Path, inputs: List[str], outputs: List[str]) -> "SklearnModel":
        serialiser = JoblibSerialiser()
        model = serialiser.deserialise(folder / cls.SKLEARN_MODEL_FILE)
        return cls(inputs=inputs, outputs=outputs, model=model, seed=1)

    @classmethod
    def get_parameters(cls: Type["SklearnModel"], parser: ArgumentParser) -> None:
        """Get and add parameters to initialise this class.

        SklearnModel's will work by requiring 2 ways to instantiate a Model:
            - through `model`, which is a sklearn model.
            - through parameters, which will instantiate the model internally.

        Args:
            parser (ArgumentParser): The ArgumentParser to add arguments to.
        """
        super().get_parameters(parser)
        BaseModel = cls._get_annotated_class()

        add_args_to_parser_for_class(
            parser, BaseModel, "model", [], excluded_args=["seed", "inputs", "outputs", "model"]
        )

    @classmethod
    def class_from_sklearn_model_class(
        cls,
        model_class: Type[SklearnModelBaseType],
    ) -> Type["SklearnModel"]:
        """Create a SklearnModel class from a SklearnModelBaseType.

        This should support all sklearn classifaction and regression models.

        Args:
            model_class (Type[SklearnModelBaseType]): The class of the sklearn model. For example,
                LinearRegression or LogisticRegression.

        Returns:
            Type[SklearnModel]: The created SklearnModel.
        """
        new_name = f"{model_class.__name__}Model"
        try:
            old_docs = cls.__doc__
            assert isinstance(old_docs, str)
        except AttributeError:
            logger.warning(f"Failed to add docstring to {new_name}.")
            old_docs = "A generic class for sklearn-like Models."

        new_doc = old_docs.replace(
            "A generic class for sklearn-like Models.",
            f"""
A generic class for sklearn-like Models.

See SklearnModel for the original source and docs for subfunctions.

This model uses a fixed class: {model_class.__name__}.
The source module is: {model_class.__module__}

""",
        )
        klass = types.new_class(
            new_name,
            (SklearnModel[model_class],),  # type: ignore
            exec_body=lambda ns: ns.update(
                {
                    "__doc__": new_doc,
                    "__module__": SklearnModel.__module__,
                }
            ),
        )
        # klass.__doc__ = new_doc

        return klass

    @classmethod
    def from_sklearn_model_class(
        cls,
        model_class: Type[SklearnModelBaseType],
        inputs: List[str],
        outputs: List[str],
        seed: int = 1,
        **model_args: Any,
    ) -> "SklearnModel":
        """Create a SklearnModel from a SklearnModelBaseType.

        This should support all sklearn classifaction and regression models.

        Args:
            model_class (Type[SklearnModelBaseType]): The class of the sklearn model. For example,
                LinearRegression or LogisticRegression.
            inputs (List[str]): A list of names of input Data. This determines which Data is
                used to fit the model.
            outputs (List[str]): A list of names of output Data. This determines the names of
                output variables.
            seed (int, optional): The RNG seed to ensure reproducability.. Defaults to 1.
            **model_args (Any): Optional keyword arguments passed to the model class to instantiate a new
                model if `model` is None.

        Returns:
            SklearnModel: The created SklearnModel.
        """
        new_cls = cls.class_from_sklearn_model_class(model_class)
        return new_cls(inputs=inputs, outputs=outputs, model=None, seed=seed, **model_args)



"""A Serialiser to integrate Spark with mlpype."""
import json
from pathlib import Path
from typing import Any, List, Type, Union

import joblib

from mlpype.base.pipeline import Pipe, Pipeline
from mlpype.base.serialiser import Serialiser
from mlpype.spark.model.types import SerialisableTransformer
from mlpype.spark.pipeline.spark_pipe import SparkPipe


class SparkSerialiser(Serialiser):
    """A Serialiser to integrate Spark with mlpype.

    It is highly recommended to initialise SparkSession manually before calling this,
    with all proper configuration done beforehand. A SparkSession is needed to
    serialise / deserialise objects stored using this class.
    """

    SUB_PIPE_PREFIX = "__sub_pipe_"
    BASE_PIPE_FILE = "pipe.pkl"
    SPARK_TRANSFORMER_FILE = "spark_transformer"
    SPARK_TRANSFORMER_CLASS_FILE = "spark_transformer_class"
    STEPS_FILE = "steps.json"

    def serialise(self, object: Any, file: Union[str, Path]) -> None:
        """Serialse a given object.

        If the object is:
            - A mlpype Pipeline: we serialise each step separately.
            - Otherwise, we use joblib.

        Args:
            object (Any): The object to serialise.
            file (Union[str, Path]): The path to serialise to.
        """
        if isinstance(object, Pipeline):
            self._serialise_pipeline(object, file)
        else:
            return self._serialise_joblib(object, file)

    def deserialise(self, file: Union[str, Path]) -> Any:
        """Deserialise the object in the given file.

        This function can handle mlpype Pipelines with PySpark elements.

        Args:
            file (Union[str, Path]): The file containing a python object to deserialise.

        Returns:
            Any: The deserialised object.
        """
        file = Path(file)
        if file.is_dir() and (file / self.STEPS_FILE).is_file():
            return self._deserialise_pipeline(file)
        else:
            return self._deserialise_joblib(file)

    def _serialise_pipeline(self, pipeline: Pipeline, file: Union[str, Path]) -> None:
        file = Path(file)

        file.mkdir(exist_ok=True)
        steps = []
        i = 0
        for step in pipeline:
            if isinstance(step, SparkPipe):
                name = step.name
                pipe_path = file / name
                pipe_path.mkdir(exist_ok=True)
                self._serialise_joblib(step, pipe_path / self.BASE_PIPE_FILE)

                # store the class for re-importing
                transformer: SerialisableTransformer = step.fitted
                self._serialise_joblib(type(transformer), pipe_path / self.SPARK_TRANSFORMER_CLASS_FILE)
                transformer.save(str(pipe_path / self.SPARK_TRANSFORMER_FILE))
            elif isinstance(step, Pipe):
                name = step.name
                self._serialise_joblib(step, file / name)
            else:
                name = f"{self.SUB_PIPE_PREFIX}{i}"
                i += 1
                self._serialise_pipeline(step, file / name)

            steps.append(name)

        with open(file / self.STEPS_FILE, "w") as f:
            json.dump({"steps": steps}, f)

    def _serialise_joblib(self, object: Any, file: Union[str, Path]) -> None:
        """Serialise the given object to the given file.

        Args:
            object (Any): The object to serialise.
            file (Union[str, Path]): The file to serialise to.
        """
        joblib.dump(object, file)

    def _deserialise_pipeline(self, file: Union[str, Path]) -> Pipeline:
        file = Path(file)
        assert file.is_dir()

        with open(file / self.STEPS_FILE, "r") as f:
            steps: List[str] = json.load(f)["steps"]

        pipes = []
        for step in steps:
            step_path = file / step
            if step_path.is_file():
                pipes.append(self._deserialise_joblib(step_path))
            elif step_path.is_dir():
                if step.startswith(self.SUB_PIPE_PREFIX):
                    pipes.append(self._deserialise_pipeline(step_path))
                else:
                    pipe: SparkPipe = self._deserialise_joblib(step_path / self.BASE_PIPE_FILE)
                    fitted_class: Type[SerialisableTransformer] = self._deserialise_joblib(
                        step_path / self.SPARK_TRANSFORMER_CLASS_FILE
                    )
                    pipe.fitted = fitted_class.load(str(step_path / self.SPARK_TRANSFORMER_FILE))
                    pipes.append(pipe)
            else:
                raise ValueError(f"{step_path} was not an existing path")

        return Pipeline(pipes)

    def _deserialise_joblib(self, file: Union[str, Path]) -> Any:
        """Deserialise the object in the given file.

        Args:
            file (Union[str, Path]): The file to deserialise.

        Returns:
            Any: The python object stored in the file.
        """
        return joblib.load(file)



"""Provides tools for using MLpype with MLflow.

At the moment, MLflow is the recommended logging tool for MLpype. This package provides a few tools:

- An MlflowLogger: An MLpype logger that uses MLflow. Set this up and you don't have to worry about how to\
    log your model.
- A quick function to load an Inferencer from an experiment stored in MLflow. You can find this in
    `mlpype.mlflow.deploy`.
- A mlflow compatible version of a Model that should work with any MLpype model. It extends mlflow's PythonModel.
    This should allow you to load any model that has been stored in MLflow using mlflow's logic. You don't
    need to do anything to use this except use the MLflowLogger. Once registered, the model can loaded using
    mlflow's `from mlflow.pyfunc import load_model`.
"""
from . import deploy, logger



"""Provides tools to simplify making Shapley plots in mlpype."""
from logging import getLogger
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from shap import Explainer, partial_dependence_plot, plots

from mlpype.base.data.dataset import DataSet
from mlpype.base.evaluate.plot import BasePlotter
from mlpype.base.experiment.experiment import Experiment
from mlpype.base.model.model import Model
from mlpype.sklearn.data.sklearn_data import SklearnData

SkleanData = Union[np.ndarray, pd.DataFrame]


class ShapleyPlot(BasePlotter):
    """Generates Shapley plots for the a model and dataset."""

    def __init__(
        self,
        input_name: str,
        output_name: str,
        sample_size: int = 100,
        sample_function: Optional[Callable[[Any, int], SkleanData]] = None,
    ):
        """
        A plotter for shapley values.

        This creates:
            - Partial dependence plots for each feature.
            - Beeswarm plots for the dataset.

        This will be run on a sampled dataset. Currently, only 1 input/output dataset is supported. By
        default, Shapley only works for numpy arrays and pandas dataframes. We do not impose this
        restriction, and leave it to the user to try and make it work for other types of data.

        Args:
            input_name (str): The name of the input dataset.
            output_name (str): The name of the output dataset.
            sample_size (int, optional): The size of the sampled dataset. Shapley values are expensive to
                compute, and can be slow. To minimise the impact of this, we sample the dataset.
                If the dataset is larger than the sample size, we use the full dataset. Defaults to 100.
            sample_function (Optional[Callable[[Any, int], SkleanData]], optional): A function to
                subsample the input dataset. By default, `_sample_df_or_np` of this object is used.
                This only supports numpy arrays and pandas dataframes.
        """
        super().__init__()
        self.input_name = input_name
        self.output_name = output_name
        self.logger = getLogger(__name__)

        if sample_function is None:
            sample_function = self._sample_df_or_np

        self.sample_function = sample_function
        self.sample_size = sample_size

    def _sample_df_or_np(self, data: SkleanData, sample_size: int) -> SkleanData:
        if isinstance(data, np.ndarray):
            if data.shape[0] >= sample_size:
                return data[np.random.choice(data.shape[0], sample_size)]
        elif isinstance(data, pd.DataFrame):
            if data.shape[0] >= sample_size:
                return data.sample(n=sample_size)
        else:
            raise ValueError(f"Input data must be either a numpy array or a pandas dataframe. Got {type(data)}")

        self.logger.warning(f"Not enough data to sample. Using full dataset. Got {data.shape[0]}")
        return data

    def plot(self, plot_folder: Path, data: DataSet, experiment: Experiment) -> List[Path]:
        """Creates plots and writes them to `plot_folder / "shapley"`.

        Args:
            plot_folder (Path): The folder to write the plot to. The final plot will be written to
                `plot_folder / "shapley"`.
            data (DataSet): A DataSet containing all the data you need to make your plots.
                In an Experiment, this will contain the last DataSet from the Pipeline with the
                predictions added as "{output_name}{Constants.PREDICTION_POSTFIX}"
            experiment (Experiment): The experiment object. Used to extract the model.

        Returns:
            List[Path]: The file paths of where the plot(s) are stored.
        """
        root_folder = plot_folder / "shapley"
        root_folder.mkdir(parents=True, exist_ok=True)

        forecast = self._make_forecast_function(experiment.model)
        sampled, shap_values = self._sample_and_shapley(data, forecast)

        result = []
        result += self._partial_plots(root_folder, forecast, sampled)
        result.append(self._beeswarm_plot(root_folder, shap_values))

        return result

    def _make_forecast_function(self, model: Model) -> Callable[[SklearnData], Any]:
        def forecast(X: SkleanData) -> SkleanData:
            return model.transform(DataSet(x=X))[self.output_name]

        return forecast

    def _sample_and_shapley(self, data: DataSet, model: Callable) -> Tuple[SkleanData, np.ndarray]:
        """
        Samples the input data and returns the shapley values.

        Args:
            data (DataSet): The input Dataset.
            model (Callable): The model-predict function.

        Returns:
            Tuple[SkleanData, np.ndarray]: The sampled data and the shapley values.
        """
        input_data = data.get(self.input_name)
        if not isinstance(input_data, (np.ndarray, pd.DataFrame)):
            self.logger.warning("Input data is not an array or dataframe. Shapley may not work.")

        # Shapley values on sampled data
        sampled = self.sample_function(input_data, self.sample_size)
        expl = Explainer(model, sampled)
        shap_values = expl(sampled)

        return sampled, shap_values

    def _partial_plots(self, root_folder: Path, model: Callable, sampled: SkleanData) -> List[Path]:
        """Plots partial dependency plots for all column."""
        if isinstance(sampled, np.ndarray):
            column_ids = list(range(sampled.shape[1]))
        elif isinstance(sampled, pd.DataFrame):
            column_ids = list(sampled.columns)
        else:
            raise ValueError(f"Input data must be either a numpy array or a pandas dataframe. Got {type(sampled)}")

        partial_deps_folder = root_folder / "partial_dependence"
        partial_deps_folder.mkdir(parents=True, exist_ok=True)

        result = []
        for col in column_ids:
            self.logger.info(f"Plotting partial dependence for column: {col}")
            partial_plot_file = partial_deps_folder / f"{col}.png"
            partial_dependence_plot(col, model, sampled, show=False)
            plt.tight_layout()
            plt.savefig(partial_plot_file)
            plt.close()
            result.append(partial_plot_file)
        return result

    def _beeswarm_plot(self, root_folder: Path, shap_values: np.ndarray) -> Path:
        """Makes a beeswarm plot."""
        self.logger.info("Plotting beeswarm plot")
        beeplot_file = root_folder / "beeswarm.png"

        plots.beeswarm(shap_values, show=False)
        plt.tight_layout()
        plt.savefig(beeplot_file)
        plt.close()

        return beeplot_file



"""Provides an implementation of the XGBoost models for mlpype."""
from pathlib import Path
from typing import List, Type, TypeVar

from xgboost.sklearn import XGBModel as BaseXGBModel

from mlpype.sklearn.model.sklearn_model import SklearnModel

T = TypeVar("T", bound=BaseXGBModel)


class XGBModel(SklearnModel[T]):
    """Provides an implementation of the XGBoost models for mlpype.

    This should be generic enough to support any XGBoost model. It extends the SklearnModel
    with specific XGBoost functionality for saving and loading models. If you want to
    use a specific (currently not already created xgboost model), just use the following example:

    ```python
    from mlpype.xgboost.model import XGBModel
    from xgboost import XGBClassifier
    XGBClassifierModel = XGBModel.class_from_sklearn_model_class(XGBClassifier)
    ```

    """

    XGB_MODEL_FILE = "model.txt"

    def _save(self, folder: Path) -> None:
        assert isinstance(self.model, BaseXGBModel)
        self.model.save_model(folder / self.XGB_MODEL_FILE)

    @classmethod
    def _load(cls: Type["XGBModel"], folder: Path, inputs: List[str], outputs: List[str]) -> "SklearnModel":
        model: T = cls._get_annotated_class()()
        model.load_model(folder / cls.XGB_MODEL_FILE)
        return cls(inputs=inputs, outputs=outputs, model=model, seed=1)



"""Provides tools to simplify plotting using matplotlib in mlpype."""
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Protocol, Union

import matplotlib.pyplot as plt

from mlpype.base.data.dataset import DataSet
from mlpype.base.evaluate.plot import BasePlotter
from mlpype.base.experiment.experiment import Experiment


class MatplotFunction(Protocol):
    """Protocol for creating plots using matplotlib."""

    def __call__(self, *data: Any) -> None:
        """Creates a plot from the given data.

        Saving the plot isn't handled by this function.

        Args:
            *data (Any): Datasets to be used in plotting.
        """


@dataclass
class MatplotlibPlotter(BasePlotter):
    """Creates plots using matplotlib."""

    plot_function: MatplotFunction
    file_name: Union[Path, str]
    dataset_names: List[str]

    def plot(self, plot_folder: Path, data: DataSet, experiment: Experiment) -> List[Path]:
        """Creates 1 plot from the given dataset and writes it to the given path using matplotlib.

        This is a simple wrapper to make a single matplotlib plot.

        Args:
            plot_folder (Path): The folder to write the plot to. The final plot will be written to
                `plot_folder / self.file_name`.
            data (DataSet): A DataSet containing all the data you need to make your plots.
                In an Experiment, this will contain the last DataSet from the Pipeline with the
                predictions added as "{output_name}{Constants.PREDICTION_POSTFIX}"
            experiment (Experiment): The experiment object. Not used by default for this
                plotter; it assumens you only use the transformed input data.

        Returns:
            List[Path]: The file paths of where the plot(s) are stored.
        """
        plot_path = plot_folder / self.file_name
        self.plot_function(*data.get_all(self.dataset_names))
        plt.savefig(plot_path)
        plt.close()
        return [plot_path]



"""An integration for hyperopt with MLpype.

This package aims to use the Experiment setup of MLpype to perform hyperparameter optimisation.

We expose 1 function, `optimise_experiment`, which can be used to perform hyperparameter \
    optimisation given an Experiment.
"""
from .optimise import optimise_experiment



"""Provides an implementation of the XGBoost model for mlpype."""
from xgboost.sklearn import XGBClassifier, XGBRanker, XGBRegressor

from .xgboost_model import XGBModel

XGBClassifierModel = XGBModel.class_from_sklearn_model_class(XGBClassifier)
XGBRegressorModel = XGBModel.class_from_sklearn_model_class(XGBRegressor)
XGBRankerModel = XGBModel.class_from_sklearn_model_class(XGBRanker)


__all__ = [
    "XGBClassifierModel",
    "XGBRegressorModel",
    "XGBRankerModel",
]



"""A mlpype-compliant framework for using Spark Models."""
import shutil
import tarfile
import typing
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, Generic, Iterable, List, Optional, Type, TypeVar, Union

from pyspark.sql import DataFrame as SparkDataFrame

from mlpype.base.data import DataSet
from mlpype.base.experiment.argument_parsing import add_args_to_parser_for_class
from mlpype.base.model import Model
from mlpype.base.serialiser import JoblibSerialiser
from mlpype.spark.model.types import SerialisablePredictor, SerialisableSparkModel

T = TypeVar("T", bound=SerialisablePredictor)
logger = getLogger(__name__)


class SparkModel(Model[SparkDataFrame], ABC, Generic[T]):
    """A mlpype-compliant framework for using Spark Models."""

    SPARK_PREDICTOR_PATH = "predictor"
    SPARK_MODEL_PATH = "spark_model"
    SPARK_MODEL_CLASS_PATH = "spark_model_class"

    def __init__(
        self,
        inputs: List[str],
        outputs: List[str],
        model: Optional[T] = None,
        predictor: Optional[T] = None,
        seed: int = 1,
        **model_args: Any,
    ) -> None:
        """A mlpype-compliant framework for using Spark Models.

        Args:
            inputs (List[str]): The name of the input DataFrame. Should be a list of 1 string.
            outputs (List[str]): The name of the output DataFrame. Should be a list of 1 string, the same as `inputs`
            model (Optional[T]): The Spark Model. Defaults to None. If set to None,
                this model can't be serialised or used for inference.
            predictor (Optional[T]): The Spark Predictor. If not set, we try to instantiate it
                using `model_args`
            seed (int, optional): Spark Seed. Currently ignored, unfortunately. Defaults to 1.
            **model_args (Any): any keywords arguments to be passed to _init_model.
        """
        assert len(inputs) == 1, "SparkML only requires a single DataFrame as input and output, the same one."
        assert len(outputs) == 1, "SparkML only requires a single DataFrame as input and output, the same one."
        assert inputs == outputs, "SparkML only requires a single DataFrame as input and output, the same one."
        super().__init__(inputs, outputs, seed)
        if predictor is None:
            predictor = self._init_model(model_args)

        self.model: Optional[SerialisableSparkModel] = model
        self.predictor = predictor

    @abstractmethod
    def _init_model(self, args: Dict[str, Any]) -> T:
        raise NotImplementedError

    @classmethod
    def _get_annotated_class(cls) -> Type[T]:
        return typing.get_args(cls.__orig_bases__[0])[0]

    def set_seed(self) -> None:
        """Sets the RNG seed."""
        # TODO: set seed properly in pyspark
        logger.warning("Currently, setting a seed is not yet supported.")

    def fit(self, data: DataSet) -> "Model":
        """Fits the Model to the given DataSet.

        The DataSet should contain all inputs. For Spark models, this means 1 DataFrame
        that contains a feature vector and the label.

        Args:
            data (DataSet): The DataSet to fit this Model on.

        Returns:
            Model: self
        """
        self._fit(*data.get_all(self.inputs))
        return self

    def _fit(self, *data: SparkDataFrame) -> None:
        assert len(data) == 1, f"SparkML needs a single DataFrame as input, got {len(data)}"
        self.model = self.predictor.fit(data[0])

    def _transform(
        self,
        *data: SparkDataFrame,
    ) -> Union[Iterable[SparkDataFrame], SparkDataFrame]:
        assert len(data) == 1, f"SparkML needs a single DataFrame as input, got {len(data)}"
        assert self.model is not None, "Please fit this model before transforming data."

        result = self.model.transform(data[0])
        return result.drop(self.model.getOrDefault("featuresCol"))

    @classmethod
    def _get_spark_prefix(cls, folder: Path) -> str:
        if folder.is_absolute():
            spark_prefix = "file://"
        else:
            logger.warning(
                "Loading model from a relative path. This may not work as expected, and will fail in Databricks!"
            )
            spark_prefix = ""
        return spark_prefix

    def _save(self, folder: Path) -> None:
        assert self.model is not None, "Please fit this model before transforming data."
        spark_prefix = self._get_spark_prefix(folder)

        serialiser = JoblibSerialiser()

        self.predictor.write().overwrite().save(spark_prefix + str(folder / self.SPARK_PREDICTOR_PATH))
        self.model.write().overwrite().save(spark_prefix + str(folder / self.SPARK_MODEL_PATH))
        serialiser.serialise(type(self.model), str(folder / self.SPARK_MODEL_CLASS_PATH))

        self._create_zip(folder / self.SPARK_PREDICTOR_PATH)
        self._create_zip(folder / self.SPARK_MODEL_PATH)

    def _create_zip(self, folder: Path) -> None:
        """Creates a zip of the serialised model or predictor."""
        folder_name = folder.name
        logger.info(f"Turning {folder_name} into a tar file")
        # assert folder.is_dir(), f"Folder {folder} does not exist."

        with tarfile.open(folder.parent / f"{folder_name}.tar", "w") as tar_ref:
            tar_ref.add(folder, arcname=folder_name)

        # Cleanup: remove input folder.
        shutil.rmtree(str(folder), ignore_errors=True)

    @classmethod
    def _load(cls: Type["SparkModel"], folder: Path, inputs: List[str], outputs: List[str]) -> "SparkModel":
        serialiser = JoblibSerialiser()
        spark_prefix = cls._get_spark_prefix(folder)

        predictor_class: Type[SerialisablePredictor] = cls._get_annotated_class()
        model_class: Type[SerialisableSparkModel] = serialiser.deserialise(str(folder / cls.SPARK_MODEL_CLASS_PATH))

        cls._unzip(folder / f"{cls.SPARK_PREDICTOR_PATH}.tar")
        cls._unzip(folder / f"{cls.SPARK_MODEL_PATH}.tar")

        predictor: SerialisablePredictor = predictor_class.load(spark_prefix + str(folder / cls.SPARK_PREDICTOR_PATH))
        model: SerialisableSparkModel = model_class().load(spark_prefix + str(folder / cls.SPARK_MODEL_PATH))

        return cls(inputs=inputs, outputs=outputs, predictor=predictor, model=model, seed=1)

    @classmethod
    def _unzip(cls, tar_file: Path) -> None:
        """Unzips the model or predictor."""
        logger.info(f"Untarring: {tar_file}")
        with tarfile.open(tar_file, "r") as tar_ref:
            tar_ref.extractall(tar_file.parent)

    @classmethod
    def get_parameters(cls: Type["SparkModel"], parser: ArgumentParser) -> None:
        """Get and add parameters to initialise this class.

        SparkModel's will work by requiring 2 ways to instantiate a Model:
            - through `model`, which is a spark model.
            - through parameters, which will instantiate the model internally.

        Args:
            parser (ArgumentParser): The ArgumentParser to add arguments to.
        """
        super().get_parameters(parser)
        BaseModel = cls._get_annotated_class()

        add_args_to_parser_for_class(
            parser,
            BaseModel,
            "model",
            [],
            excluded_args=["seed", "inputs", "outputs", "model"],
        )



"""Provides classes for sklearn models.

For sklearn models not already configured here, you can use the SklearnModel class to quickly incorporate your model.
We have already integrated classifier and regressor models from sklearn. You can import them like:
```python
from mlpype.sklearn.model import <sklearn name>Model
```
"""

#  This weird import setup is required since sklearn.utils seems to want to refer to mlpype.sklearn.utils
import importlib
import sys

from .sklearn_base_type import SklearnModelBaseType
from .sklearn_model import SklearnModel

# from sklearn.linear_model import LinearRegression, LogisticRegression


__utils = importlib.import_module("sklearn.utils", package=None)
__classes = __utils.all_estimators(["classifier", "regressor"])
# classes = [LinearRegression, LogisticRegression]

__dynamic_classes = []
for name, klass in __classes:
    __new_name = f"{name}Model"
    __dynamic_class = SklearnModel.class_from_sklearn_model_class(klass)
    __dynamic_classes.append(__new_name)
    setattr(sys.modules[__name__], __new_name, __dynamic_class)

__all__ = ["SklearnModel", "SklearnModelBaseType", *__dynamic_classes]



"""Provides tools for guaranteeing that SparkSession is active."""
from typing import Optional

from pyspark.sql import SparkSession


def guarantee_spark(
    spark_session: Optional[SparkSession] = None,
) -> SparkSession:
    """Checks if the given spark_session object is None and if so, returns the active session instead.

    Args:
        spark_session (Optional[SparkSession]): The spark_session object to guarantee. Defaults to None.

    Returns:
        SparkSession: The active SparkSession.

    Raises:
        AssertionError: If no SparkSession is active.
    """
    if spark_session is None:
        spark_session = SparkSession.getActiveSession()
    assert isinstance(spark_session, SparkSession), "No active SparkSession"
    return spark_session



"""Provides tools for using Spark with MLpype.

Important to note is that serialisation and evaluation work differently in Spark.

- Evaluation works slightly different, not taking a dictionary of functions but a JavaEvaluator.
- Serialisation has to take the SparkSession into account and serialise the trained Pipeline and Model properly.
"""
from . import data, evaluate, model, pipeline, serialiser



"""Provides a logger using mlflow for mlpype."""
import logging
from pathlib import Path
from types import TracebackType
from typing import Any, Dict, Optional, Type, Union

from git import InvalidGitRepositoryError
from git.repo import Repo
from mlflow import create_experiment  # type: ignore
from mlflow import get_experiment_by_name  # type: ignore
from mlflow import log_artifact  # type: ignore
from mlflow import log_metrics  # type: ignore
from mlflow import log_params  # type: ignore
from mlflow import register_model  # type: ignore
from mlflow import set_experiment  # type: ignore
from mlflow import set_tag  # type: ignore
from mlflow import set_tracking_uri  # type: ignore
from mlflow import start_run  # type: ignore
from mlflow.pyfunc import log_model as mlflow_log_model  # type: ignore

from mlpype.base.logger import ExperimentLogger
from mlpype.base.model.model import Model
from mlpype.mlflow.model.model import PypeMLFlowModel


class MlflowLogger(ExperimentLogger):
    """A logger using mlflow for mlpype."""

    ARTIFACT_FOLDER = "artifact_folder"

    def __init__(
        self, name: str, uri: str, artifact_location: Optional[str] = None, model_name: Optional[str] = None
    ) -> None:
        """A logger using mlflow for mlpype.

        Args:
            name (str): The name of the experiment. For runs done on databricks, make sure you preface
                this with a workspace, like /Users/<your databricks user name>/<experiment name>.
            uri (str): The tracking uri. Should most likely start with `http://` or `databricks`.
                It can also equal `databricks`.
            artifact_location (Optional[str]): The artificat uri location. This is needed if you set up
                your own artifact location, to make sure files are copied over to any remote tracking server.
                This is not required when using hosted databricks. Defaults to None.
            model_name (Optional[str]): The name to register the model under. This will register the results
                of your experiment as a model in mlflow. If set to None, your model will not be registered.
        """
        super().__init__()
        logger = logging.getLogger(__name__)
        self.name = name

        if not uri.startswith("http://") and not uri.startswith("databricks"):
            logger.warning(f"Most often `uri` has to start with `http://` or `databricks`. Got: {uri}")
        self.uri = uri
        self.artifact_location = artifact_location
        self.run = None
        self.model_name = model_name

    def __enter__(self) -> None:
        """Setup the mlflow experiment and call start_run."""
        set_tracking_uri(self.uri)
        experiment = get_experiment_by_name(self.name)

        if experiment is None:
            # this is a new experiment: create it explicitly with the artifact uri
            # TODO: make sure this works on S3
            create_experiment(self.name, self.artifact_location)

        set_experiment(experiment_name=self.name)

        self.run = start_run().__enter__()
        self.log_branch()

    def log_branch(self) -> None:
        """Set the current branch as a tag on the experiment."""
        assert self.run is not None, "Please start the experiment first"
        repo = self._find_encapsulating_repo(Path("."))

        if repo is None:
            branch_name = "unknown_branch"
        else:
            branch_name = repo.active_branch.name

        set_tag("git_branch", branch_name)

    def _find_encapsulating_repo(self, directory: Path) -> Union[Repo, None]:
        """Finds the current repo we're in, if any.

        We'll recursively move up from `directory` until we find a valid
        git repo.

        Args:
            directory (Path): The directory to start looking for. Usually set
                to the current directory.

        Returns:
            Union[Repo, None]: None if no repo is found, otherwise a GitPython Repo.
        """
        repo = None
        directory = directory.absolute()
        while str(directory) != "/":
            try:
                repo = Repo(directory)
                break
            except InvalidGitRepositoryError:
                directory = directory.parent
        return repo

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        """Exits a run started by __enter__.

        Args:
            exc_type (Optional[Type[BaseException]]): Passed to ActiveRun.__exit__()
            exc_val (Optional[BaseException]): Passed to ActiveRun.__exit__()
            exc_tb (Optional[TracebackType]): Passed to ActiveRun.__exit__()
        """
        assert self.run is not None, "Run not started, cannot exit!"
        self.run.__exit__(exc_type, exc_val, exc_tb)

    def _log_metrics(self, metrics: Dict[str, Union[str, float, int, str, bool]]) -> None:
        """Perform the actual logging of metrics in mlflow.

        Args:
            metrics (Dict[str, Union[str, float, int, str, bool]]): A dictionary of metric names and values.
        """
        log_metrics(metrics)

    def log_parameters(self, parameters: Dict[str, Any]) -> None:
        """Logs the parameters for a given run.

        These will by default be passed on with prefixes such as `model__`.

        Args:
            parameters (Dict[str, Any]): The parameters of a given run. Ideally these
                parameters should be no more complicated than string, float, int, bool, or a
                list of these.
        """
        log_params(parameters)

    def log_model(self, model: Model, folder: Union[str, Path]) -> None:
        """Logs a Model for a given experiment.

        This function will write the model to the given location as well.
        This specific version for MLflow also uses their logging format.

        Please note that if you want to register this model in MLflow, you
        can use "runs:/{run_id}/{MlflowLogger.ARTIFACT_FOLDER}" to refer to the MLflow model.

        Args:
            model (Model): The Model to be logged.
            folder (Union[str, Path]): The file to log the Model to.
        """
        folder = Path(folder)
        model.save(folder)
        super().log_model(model, folder)
        mlflow_log_model(
            artifact_path=self.ARTIFACT_FOLDER,
            # PypeMLFlowModel is a near-dummy class. It knows how to load a model
            # from a MLpype training result.
            python_model=PypeMLFlowModel(),
            artifacts={"folder": str(folder.parent)},
        )

        if self.model_name is not None:
            assert self.run is not None, "Please start the experiment first"
            self.register_mlpype_model(self.run.info.run_id, self.model_name)

    def log_file(self, file: Union[str, Path]) -> None:
        """Logs a given file as part of an experiment.

        This only works well if:
            - you work with a databricks-hosted mlflow experiment.
            - you have set a artifact uri and configured your server properly.
            - you have run mlflow ui/server at the same folder where mlruns is located.

        Args:
            file (Union[str, Path]): The file to log.
        """
        file = Path(file)
        log_artifact(str(file))

    def register_mlpype_model(self, run_id: Optional[str], model_name: Optional[str]) -> None:
        """Applies the register_model function of mlflow to MLpype-trained models.

        Specifically, this registers the artifact folder that we created
        in the log_model function.

        Args:
            run_id (Optional[str]): The run id of the model. If not provided, it will be taken from
                self.run
            model_name (Optional[str]): The name to register the model under (using register_model).
                If not provided, it will be taken from self.model_name, assuming it's set.
        """
        if run_id is None:
            assert self.run is not None, "Please start the experiment first"
            run_id = self.run.info.run_id
        if model_name is None:
            assert self.model_name is not None, "Please provide a model name"
            model_name = self.model_name
        register_model(f"runs:/{run_id}/{self.ARTIFACT_FOLDER}", model_name)



"""Provides tools to simplify plotting using matplotlib in mlpype.

This allows you to create matplotlib plots in mlpype. They will be stored in the experiment for easy access.
"""
from . import evaluate

__all__ = ["evaluate"]
